
// AudioRecorder.java
// Andrew Davison, ad@fivedots.coe.psu.ac.th, Feb. 2022

// http://jsresources.sourceforge.net/examples/AudioRecorder.java.html
// http://www.java2s.com/Code/Java/Development-Class/CapturingAudiowithJavaSoundAPI.htm
// https://www.baeldung.com/java-sound-api-capture-mic

/* Record audio to a WAV file

*/


import java.io.*;
import java.util.*;
import javax.sound.sampled.*;



public class AudioRecorder extends Thread
{
  // The sample rate: 44,100 Hz for CD quality audio
  public static final int SAMPLE_RATE = 44100;
  private static final int BITS_PER_SAMPLE = 16;         // 16-bit audio
  private static final int MONO = 1;
  private static final boolean LITTLE_ENDIAN = false;
  private static final boolean SIGNED = true;


  private TargetDataLine line;
  private AudioFormat audioFormat;
  private String outFnm;
  private boolean isRecording = false;


  public AudioRecorder(String fnm)
  {
    outFnm = fnm;

    // connect data line to microphone
    audioFormat = new AudioFormat((float) SAMPLE_RATE, BITS_PER_SAMPLE,
                                   MONO, SIGNED, LITTLE_ENDIAN);
    line = AudioTools.getLine(audioFormat);
    System.out.println("Connected to a microphone...");
  }  // end of AudioRecorder()



  public void start()
  {
    line.start();
    super.start();   // start thread
   } 


  public void stopRecording()
  {
    // for recording, the line needs to be stopped before draining 
    line.stop();
    line.drain();
    line.close();
    isRecording = false;
  } /* stopRecording */


  public void run()
  {
    // target data line --> buffer --> baos
    ByteArrayOutputStream baos = new ByteArrayOutputStream();
    int bufferSize = (int) audioFormat.getSampleRate() * audioFormat.getFrameSize();
    byte[]  buffer = new byte[bufferSize];
    isRecording = true;

    while (isRecording) {
      int count = line.read(buffer, 0, buffer.length);
      if (count > 0)
        baos.write(buffer, 0, count);
    }
    try {
      baos.close();
    }
    catch (IOException e)
    {  System.out.println("Error: could not close microphone input");  }

    // baos --> data --> bais --> ais --> WAV file
    byte[] data = baos.toByteArray();
    ByteArrayInputStream bais = new ByteArrayInputStream(data);
    AudioInputStream ais = new AudioInputStream(bais, audioFormat,
                                   data.length / audioFormat.getFrameSize());

    long ms = (long) ((ais.getFrameLength() * 1000) / audioFormat.getFrameRate());
    double duration = (double)ms/1000.0;
    System.out.printf("Recorded duration: %.2f secs\n", duration);

    AudioTools.saveAudio(outFnm, ais);
  } // end of run()


  // ---------------------------------------------------

  public static void main(String[] args)
  {
    if (args.length != 1) {
      System.out.println("Usage: java AudioRecorder <outfnm>");
      return;
    }

    AudioRecorder recorder = new AudioRecorder(args[0]);

    Scanner scan = new Scanner(System.in);
    System.out.println("Press ENTER to start recording...");
    scan.nextLine();
    recorder.start();

    System.out.println("Recording. Press ENTER to stop...");
    scan.nextLine();
    recorder.stopRecording();
    System.out.println("Recording stopped.");
  }    // end of main()



}  // end of AudioRecorder class
