
// AudioTools.java
// Andrew Davison, ad@fivedots.coe.psu.ac.th, Feb. 2022

/* This class provides basic capabilities for
   creating, reading, playing, and saving audio.

   It is primarily a combination of StdAudio.java by
   Robert Sedgewick and Kevin Wayne
   (see https://introcs.cs.princeton.edu/java/stdlib/StdAudio.java.html)

   and AudioEffects.java from
   https://katie.cs.mtech.edu/classes/archive/f16/csci135/assign/audio/

   Other useful resources:
      http://jsresources.sourceforge.net/
      http://www.java2s.com/Code/Java/Development-Class/Sound.htm
      http://www.java2s.com/Code/Java/Development-Class/DukeSpeaksTest.htm

   -----------------
   Functions:

       -- play audio ---
       void play(String fnm)
       void playBG(String fnm)
       void loop(String fnm)
       void loopBG(String fnm)

       void play(double[] a)

       -- load and save ---
       double[] read(String fnm)
       Clip loadClip(String fnm)
       void save(String fnm, double[] a)
       void saveArray(String fnm, double[] audio)
       void close()

       --- recording audio ---
       TargetDataLine getLine(AudioFormat format) 
       void saveAudio(String fnm, AudioInputStream ais) 

       --- tone generation ---
       double[] tone(String key, double duration)
       double[] tone(String key, double duration, double amplitude)
       double[] tone(double freq, double duration)
       double[] tone(double freq, double duration, double amplitude)

       double[] loadTones(String fnm)

       double getFrequency(String key)
       double keyNum2Freq(int keyNumber)

       double[] harmonic(String key, double duration)

       --- wave functions ---
       double sinWave(double freq, int i)
       double sawWave(double freq, int i)
       double triangleWave(double freq, int i)
       double squareWave(double freq, int i)

       --- others ---
       void pause(int ms)
       void stats(double[] audio)
       void draw(double[] audio)
       void mixerInfo()

       --- audio array manipulation ---
       double[] copy(double[] arr)
       double[] concat(double[]... arrs)
       double[] average(double[]... arrs)
       double[] weightSum(double[] a, double[] b, double weight)
       double[] setVolume(double[] audio, double gain)
       double[] reverse(double[] audio)
       double[] stretch(double[] audio)
       double[] half(double[] audio)
       double[] noise(double[] audio)
       double[] setLoops(double[] audio, int numLoops)
       double[] merge(double[] audio1, double[] audio2)
       double[] cut(double[] audio, double start, double end)
       double[] pad(double[] audio, double start, double end)
       double[] echo(double[] audio, double secs, double vol)


   The audio format uses a sampling rate of 44,100 Hz, 16-bit, mono aural.

   For additional documentation, see
   https://introcs.cs.princeton.edu/15inout
*/


import java.io.*;
import java.nio.file.*;
import java.awt.*;
import java.util.*;
import java.util.function.*;
import java.text.*;

import java.awt.geom.Line2D;
import java.awt.image.BufferedImage;
import javax.imageio.ImageIO;
import java.net.URL;
import javax.sound.sampled.*;


public class AudioTools
{

  // Sample per second: 44,100 Hz for CD quality audio
  public static final int SAMPLE_RATE = 44100;

  private static final int BYTES_PER_SAMPLE = 2;         // 16-bit audio
  private static final int BITS_PER_SAMPLE = 16;         // 16-bit audio
  private static final int MAX_16_BIT = 32768;
  private static final int SAMPLE_BUFFER_SIZE = 4096;

  private static final int MONO = 1;
  private static final int STEREO = 2;

  private static final boolean LITTLE_ENDIAN = false;
  private static final boolean BIG_ENDIAN = true;

  private static final boolean SIGNED = true;
  private static final boolean UNSIGNED = false;

  private static final String DRAW_FNM = "audioDrawing.png";
  private static final int IM_WIDTH = 800;
  private static final int IM_HEIGHT = 200;

  private static final int MAX_VALS_PER_LINE = 6;

  private static final double MAX_NOISE = 0.05;


  private static final Object[] notes = {   // for tone generation
    "C0", 16.35,
    "C#0/Db0", 17.32,
    "D0", 18.35,
    "D#0/Eb0", 19.45,
    "E0", 20.6,
    "F0", 21.83,
    "F#0/Gb0", 23.12,
    "G0", 24.5,
    "G#0/Ab0", 25.96,
    "A0", 27.5,
    "A#0/Bb0", 29.14,
    "B0", 30.87,

    "C1", 32.7,  // Pedal C
    "C#1/Db1", 34.65,
    "D1", 36.71,
    "D#1/Eb1", 38.89,
    "E1", 41.2,
    "F1", 43.65,
    "F#1/Gb1", 46.25,
    "G1", 49.00,
    "G#1/Ab1", 51.91,
    "A1", 55.00,
    "A#1/Bb1", 58.27,
    "B1", 61.74,

    "C2", 65.41,   // Deep C
    "C#2/Db2", 69.3,
    "D2", 73.42,
    "D#2/Eb2", 77.78,
    "E2", 82.41,
    "F2", 87.31,
    "F#2/Gb2", 92.5,
    "G2", 98.00,
    "G#2/Ab2", 103.83,
    "A2", 110.00,
    "A#2/Bb2", 116.54,
    "B2", 123.47,

    "C3", 130.81,
    "C#3/Db3", 138.59,
    "D3", 146.83,
    "D#3/Eb3", 155.56,
    "E3", 164.81,
    "F3", 174.61,
    "F#3/Gb3", 185.00,
    "G3", 196.00,
    "G#3/Ab3", 207.65,
    "A3", 220.00,
    "A#3/Bb3", 233.08,
    "B3", 246.94,

    "C4", 261.63,  // Middle C
    "C#4/Db4", 277.18,
    "D4", 293.66,
    "D#4/Eb4", 311.13,
    "E4", 329.63,
    "F4", 349.23,
    "F#4/Gb4", 369.99,
    "G4", 392.00,
    "G#4/Ab4", 415.3,
    "A4", 440.00,
    "A#4/Bb4", 466.16,
    "B4", 493.88,

    "C5", 523.25,  // Tenor C
    "C#5/Db5", 554.37,
    "D5", 587.33,
    "D#5/Eb5", 622.25,
    "E5", 659.26,
    "F5", 698.46,
    "F#5/Gb5", 739.99,
    "G5", 783.99,
    "G#5/Ab5", 830.61,
    "A5", 880.00,
    "A#5/Bb5", 932.33,
    "B5", 987.77,

    "C6", 1046.5,  // Soprano C (High C)
    "C#6/Db6", 1108.73,
    "D6", 1174.66,
    "D#6/Eb6", 1244.51,
    "E6", 1318.51,
    "F6", 1396.91,
    "F#6/Gb6", 1479.98,
    "G6", 1567.98,
    "G#6/Ab6", 1661.22,
    "A6", 1760.00,
    "A#6/Bb6", 1864.66,
    "B6", 1975.53,

    "C7", 2093.00,  // Double high C
    "C#7/Db7", 2217.46,
    "D7", 2349.32,
    "D#7/Eb7", 2489.02,
    "E7", 2637.02,
    "F7", 2793.83,
    "F#7/Gb7", 2959.96,
    "G7", 3135.96,
    "G#7/Ab7", 3322.44,
    "A7", 3520.00,
    "A#7/Bb7", 3729.31,
    "B7", 3951.07,

    "C8", 4186.01, //  Eighth octave
    "C#8/Db8", 4434.92,
    "D8", 4698.64,
    "D#8/Eb8", 4978.03,
    "E8", 5274.041,
    "F8", 5587.652,
    "F#8/Gb8", 5919.911,
    "G8", 6271.927,
    "G#8/Ab8", 6644.875,
    "A8", 7040.000,
    "A#8/Bb8", 7458.620,
    "B8", 7902.133,
  };


  private static SourceDataLine line;     // for playing sound

  private static byte[] buffer;
  private static int bufferSize = 0;
              // number of samples currently in buffer

  private static HashMap<String, Double> noteMap;
  private static Random rand;

  private AudioTools()
  {
    // can not instantiate
  }


  // static initialization -------------------

  static {
    buildNotesMap();
    rand = new Random();
    initLine();
  }


  private static void buildNotesMap()
  {
    noteMap = new HashMap<String, Double>();
    for(int i = 0; i < notes.length; i=i+2) {
      String name = (String)notes[i];
      double freq = (Double)notes[i+1];
      String[] keys = name.split("/");
      for(String key : keys) {
        noteMap.put(key, freq);
        // System.out.println(key);
      }
    }
  }  // end of buildNotesMap()



  private static void initLine()
  // open up an audio stream
  {
    try {
      // 44,100 Hz, 16-bit audio, mono, signed PCM, little endian
      AudioFormat format = new AudioFormat((float) SAMPLE_RATE,
                                           BITS_PER_SAMPLE,
                                           MONO, SIGNED, LITTLE_ENDIAN);
      DataLine.Info info = new DataLine.Info(SourceDataLine.class, format);

      line = (SourceDataLine) AudioSystem.getLine(info);
      line.open(format, SAMPLE_BUFFER_SIZE * BYTES_PER_SAMPLE);

      /* The internal buffer is a fraction of the actual buffer size,
         this choice is arbitrary.
         It gets divided because we can't expect the buffered data to
         line up exactly with when the sound card decides 
         to push out its samples.
      */
      buffer = new byte[SAMPLE_BUFFER_SIZE * BYTES_PER_SAMPLE/3];
    }
    catch (LineUnavailableException e) {
      System.out.println("Could not create an audio line: " + e.getMessage());
      System.exit(1);
    }

    // no sound gets made before this call
    line.start();
  } /* initLine */


  public static void close()
  {
    line.drain();
    line.stop();
  }


  // ------------- play audio -------------

  public static synchronized void play(String fnm)
  {  playBG(fnm); }


  public static synchronized void playBG(String fnm)
  {  playBG(fnm, null);  }


  public static synchronized void playBG(String fnm, Viewer v)
  /* Plays an audio file (in .wav, .mid, or .au format) in a background thread.
     The Viewer will receive the audio data as it plays, and
     can display it. See ChartAudio.java and Viewer.java.
     An example of playing audio in this way is in TestSounds.java
  */
  {
    new Thread(new Runnable() {
      public void run() {
        AudioInputStream ais = getAudioStream(fnm);
        playAudioStream(ais, v);
      }
    }).start();
  } /* playBG */



  private static AudioInputStream getAudioStream(String fnm)
  {
    if (fnm == null) {
      System.out.println("Error: filename is null");
      return null;
    }

    System.out.println("Opening file: " + fnm);
    try {
      // try to read file from local file system
      File file = new File(fnm);
      if (file.exists())
        return AudioSystem.getAudioInputStream(file);

      // resource relative to .class file
      InputStream is1 = AudioTools.class.getResourceAsStream(fnm);
      if (is1 != null)
        return AudioSystem.getAudioInputStream(is1);

      // resource relative to classloader root
      InputStream is2 = AudioTools.class.getClassLoader().getResourceAsStream(fnm);
      if (is2 != null)
        return AudioSystem.getAudioInputStream(is2);

      // from URL (including jar file)
      URL url = new URL(fnm);
      if (url != null)
        return AudioSystem.getAudioInputStream(url);
      else  { // give up
        System.out.println("Error: could not read");
        return null;
      }
    }
    catch (IOException e) {
      System.out.println("Error: could not read");
      return null;
    }
    catch (UnsupportedAudioFileException e) {
      System.out.println("Error: unsupported audio format");
      return null;
    }
  } /* getAudioStream */




  private static void playAudioStream(AudioInputStream ais, Viewer v)
  /* https://www3.ntu.edu.sg/home/ehchua/programming/java/J8c_PlayingSound.html
     play a wav or aif file;
     javax.sound.sampled.Clip fails for long clips (on some systems),
     perhaps because JVM closes (see remedy in loop)
  */
  {
    if (ais == null)
      return;

    SourceDataLine line = null;
    int BUFFER_SIZE = 4096;     // 4K buffer
    System.out.println("Playing audio");
    try {
      AudioFormat audioFormat = ais.getFormat();
      DataLine.Info info = new DataLine.Info(SourceDataLine.class, audioFormat);
      line = (SourceDataLine) AudioSystem.getLine(info);
      line.open(audioFormat);
      line.start();
      byte[] samples = new byte[BUFFER_SIZE];
      int count = 0;
      long prevTime = System.currentTimeMillis();   // System.nanoTime();
      long currTime;
      while ((count = ais.read(samples, 0, BUFFER_SIZE)) != -1) {
        line.write(samples, 0, count);
        if (v != null) {
          currTime = System.currentTimeMillis();   // System.nanoTime();
          v.update( bytes2doubles(samples, audioFormat), currTime - prevTime);
          prevTime = currTime;
        }
      }
    }
    catch (IOException e) {
      System.out.println("Error: could not read audio samples");
    }
    catch (LineUnavailableException e) {
      System.out.println("Error: no audio line available");
    }
    finally {
      if (v != null)
        v.finished();
      if (line != null) {
        line.drain();
        line.close();
      }
    }
  } /* playAudioStream */



  public static synchronized void loop(String fnm)
  {  loopBG(fnm); }


  public static synchronized void loopBG(String fnm)
  // Loops an audio file (in .wav, .mid, or .au format) in a background thread.
  {
    if (fnm == null) {
      System.out.println("Error: filename is null");
      return;
    }

    AudioInputStream ais = getAudioStream(fnm);
    if (ais == null)
      return;

    try {
      System.out.println("Looping...");

      Clip clip = AudioSystem.getClip();
      // Clip clip = (Clip) AudioSystem.getLine(new Line.Info(Clip.class));
      clip.open(ais);
      clip.loop(Clip.LOOP_CONTINUOUSLY);
    }
    catch (LineUnavailableException e) {
      System.out.println("Error: no audio line available");
    }
    catch (IOException e) {
      System.out.println("Error: unable to read audio samples");
    }

    // keep JVM open
    new Thread(new Runnable() {
      public void run() {
        while (true)
          AudioTools.pause(1000);
      }
    }).start();
  } /* loopBG */



  public static void play(double[] samples)
  /* Writes the array of samples (between -1.0 and +1.0) to standard audio.
     If a sample is outside the range, it will be clipped.
  */
  {
    if (samples == null) {
      System.out.println("Error: audio array is null");
      return;
    }
    for (int i = 0; i < samples.length; i++)
      play(samples[i]);
  } /* play */



  private static void play(double sample)
  /* Writes one sample (between -1.0 and +1.0) to standard audio.
     If the sample is outside the range, it will be clipped.
  */
  {
    if (Double.isNaN(sample)) {
      System.out.println("Error: sample value is NaN");
      return;
    }

    // clip if outside [-1, +1]
    if (sample < -1)
      sample = -1;
    else if (sample > 1)
      sample = 1.0;

    // convert to bytes
    short s = (short) (MAX_16_BIT * sample);
    if (sample == 1.0)
      s = Short.MAX_VALUE;       // special case since 32768 not a short

    buffer[bufferSize++] = (byte) s;
    buffer[bufferSize++] = (byte) (s >> 8);       // little endian

    // send to sound card if buffer is full
    if (bufferSize >= buffer.length) {
      line.write(buffer, 0, buffer.length);
      bufferSize = 0;
    }
  } /* play */


  // ------------- load and save -------------


  public static double[] read(String fnm)
  /* Reads audio samples from a file (in .wav or .au format) and returns
     them as a double array with values between -1.0 and +1.0.
     The audio file must be 16-bit with a sampling rate of 44,100.
     It can be mono or stereo.
  */
  {
    // make sure that AudioFormat is 16-bit, 44,100 Hz, little endian
    AudioInputStream ais = getAudioStream(fnm);
    if (ais == null)
      return null;

    AudioFormat audioFormat = ais.getFormat();
    if (!isValidFormat(audioFormat))
      return null;

    System.out.println("Reading audio data from " + fnm);
    byte[] bytes = null;
    try {
      int bytesToRead = ais.available();
      bytes = new byte[bytesToRead];
      int bytesRead = ais.read(bytes);
      if (bytesToRead != bytesRead) {
        System.out.println("Error: only read " + bytesRead +
                                     " of " + bytesToRead + " bytes");
        return null;
      }
    }
    catch (IOException ioe) {
      System.out.println("Error: could not read");
      return null;
    }

    return bytes2doubles(bytes, audioFormat);
  } /* read */



  private static boolean isValidFormat(AudioFormat audioFormat)
  {
    // System.out.println("Audio Format: " + audioFormat);

    // require sampling rate = 44,100 Hz
    if (audioFormat.getSampleRate() != SAMPLE_RATE) {
      System.out.println("AudioTools only supports a sample rate of " +
                          SAMPLE_RATE + " Hz\n" +
                          "audio format: " + audioFormat);
      return false;
    }

    // require 16-bit audio
    if (audioFormat.getSampleSizeInBits() != BITS_PER_SAMPLE) {
      System.out.println("Error: AudioTools only supports " +
                         BITS_PER_SAMPLE + "-bit audio\n" +
                         "audio format: " + audioFormat);
      return false;
    }

    // require little endian
    if (audioFormat.isBigEndian()) {
      System.out.println("Error: AudioTools only supports little endian\n"
                                         + "audio format: " + audioFormat);
      return false;
    }

    return true;
  }  // end of isValidFormat()



  private static double[] bytes2doubles(byte[] bytes, AudioFormat audioFormat)
  {
    int n = bytes.length;

    // little endian, mono
    if (audioFormat.getChannels() == MONO) {
      double[] audio = new double[n/2];
      for (int i = 0; i < n/2; i++)
        // little endian, mono
        audio[i] = ((short) (((bytes[2*i+1] & 0xFF) << 8) |
                            (bytes[2*i] & 0xFF))) / ((double) MAX_16_BIT);
      return audio;
    }  
    // little endian, stereo
    else if (audioFormat.getChannels() == STEREO) {
      double[] audio = new double[n/4];
      for (int i = 0; i < n/4; i++) {
        double left  =
          ((short) (((bytes[4*i+1] & 0xFF) << 8) |
                     (bytes[4*i + 0] & 0xFF))) / ((double) MAX_16_BIT);
        double right =
          ((short) (((bytes[4*i+3] & 0xFF) << 8) |
                     (bytes[4*i + 2] & 0xFF))) / ((double) MAX_16_BIT);
        audio[i] = (left + right) / 2.0;
      }
      return audio;
    }
    else {   // TODO: handle big endian (or other formats)
      System.out.println("Error: audio format is neither mono or stereo");
      return null;
    }
  }  // end of bytes2doubles()



  public static Clip loadClip(String fnm)
  {
    Clip clip = null;

    // make sure that AudioFormat is 16-bit, 44,100 Hz, little endian
    AudioInputStream ais = getAudioStream(fnm);
    if (ais == null)
      return null;

    AudioFormat audioFormat = ais.getFormat();
    if (!isValidFormat(audioFormat))
      return null;

    System.out.println("Reading clip from " + fnm);
    try {
      DataLine.Info info = new DataLine.Info(Clip.class, audioFormat);

      // make sure sound system supports data line
      if (!AudioSystem.isLineSupported(info)) {
        System.out.println("Unsupported clip format");
        return null;
      }

      clip = (Clip) AudioSystem.getLine(info);

      clip.open(ais);    // open the sound file as a clip
      ais.close(); // we're done with the input stream
    }
    catch (LineUnavailableException noLineException) {
      System.out.println("No audio line available");
    }
    catch (IOException ioException) {
      System.out.println("Error: could not read");
    }
    catch (Exception e) {
      System.out.println("Problem with " + fnm);
    }

    return clip;
  } // end of loadClip()



  public static double[] readArray(String fnm)
  /* Reads doubles from a text file and returns
     them as a double array with values between -1.0 and +1.0.
  */
  {
    System.out.println("Reading audio data array from " + fnm);

    ArrayList<Double> vals = new ArrayList<>();
    try {
      Scanner in = new Scanner(new File(fnm));
      while(in.hasNextDouble()) {
        double d = in.nextDouble();
        if ((d >= -1.0) && (d <= 1.0))
          vals.add(d);
        else
          System.out.println("Error: data out of range");
      }
      in.close();
    } 
    catch (FileNotFoundException e) {
      System.out.println("Error: could not find file");
    }

    double[] audio = new double[vals.size()];
    for(int i = 0; i < vals.size(); i++)
      audio[i] = vals.get(i);
    return audio;
  }  // end of readArray()



  public static void save(String fnm, double[] samples)
  // Save the array as an audio file (using .wav or .au format)
  {
    if (fnm == null) {
      System.out.println("Error: filename is null");
      return;
    }

    if (samples == null) {
      System.out.println("Error: audio array is null");
      return;
    }

    // assumes 16-bit samples with sample rate = 44,100 Hz
    // use 16-bit audio, mono, signed PCM, little Endian
    AudioFormat format = new AudioFormat(SAMPLE_RATE, 16, MONO,
                                         SIGNED,  LITTLE_ENDIAN);
    byte[] data = new byte[2 * samples.length];
    for (int i = 0; i < samples.length; i++) {
      int temp = (short) (samples[i] * MAX_16_BIT);
      if (samples[i] == 1.0)
        temp = Short.MAX_VALUE;     // special case since 32768 not a short
      data[2*i + 0] = (byte) temp;
      data[2*i + 1] = (byte) (temp >> 8);         // little endian
    }

    // now save the file
    System.out.println("Saving audio to " + fnm);
    try {
      ByteArrayInputStream bais = new ByteArrayInputStream(data);
      AudioInputStream ais = new AudioInputStream(bais, format, samples.length);
      if (fnm.endsWith(".wav") || fnm.endsWith(".WAV"))
        AudioSystem.write(ais, AudioFileFormat.Type.WAVE, new File(fnm));
      else if (fnm.endsWith(".au") || fnm.endsWith(".AU"))
        AudioSystem.write(ais, AudioFileFormat.Type.AU, new File(fnm));
      else {
        System.out.println("Error: file type must be .wav or .au");
        return;
      }
    }
    catch (IOException ioe) {
      System.out.println("Error: unable to save audio");
    }
  } /* save */


  public static void saveArray(String fnm, double[] audio)
  {
    if (fnm == null) {
      System.out.println("Error: filename is null");
      return;
    }

    if (audio == null) {
      System.out.println("Error: audio array is null");
      return;
    }

    DecimalFormat df = new DecimalFormat("#.######");
    int valCount = 0;
    try {
      System.out.println("Saving audio array data to " + fnm);

      BufferedWriter bw = new BufferedWriter( new FileWriter(fnm));
      for (int i = 0; i < audio.length; i++) {
        bw.write(df.format(audio[i]) + " ");
        valCount++;
        if (valCount == MAX_VALS_PER_LINE) {
          bw.newLine();
          valCount = 0;
        }
      }
      bw.flush();  
      bw.close();  
    }
    catch (IOException ioe) {
      System.out.println("Error: unable to save audio array data");
    }
  }  // end of saveArray()


  // ------------- recording audio ---------------


  public static TargetDataLine getLine(AudioFormat format) 
  {
    TargetDataLine line = null;
    DataLine.Info info = new DataLine.Info(TargetDataLine.class, format);
    if (!AudioSystem.isLineSupported(info)) {
      System.out.println("Error: line not supported");
      System.exit(1);
    }

    try {
      line = (TargetDataLine) AudioSystem.getLine(info);
      line.open(format);
    } 
    catch (Exception ex) {
      System.out.println("Error: Could not create an audio line");
      System.exit(1);
    }

    if (line == null) {
      System.out.println("Error: can't get TargetDataLine");
      System.exit(1);
    }
    return line;
  }  // end of getLine()



  public static void saveAudio(String fnm, AudioInputStream ais) 
  {
    if (ais == null) {
      System.out.println("Error: audio stream is null");
      return;
    }

    try {
      ais.reset();
    } 
    catch (Exception e) {
      System.out.println("Error: audio stream could not be rewound");
      return;
    }

    System.out.println("Saving audio stream to file: " + fnm);
    try {
      AudioSystem.write(ais, AudioFileFormat.Type.WAVE, new File(fnm));
    } 
    catch (Exception ex) {
      System.out.println("Error: could not write audio");
      return;
    }
  }  // end of saveAudio()



  // -------------- tone generation ---------------------



  public static double[] tone(String key, double duration)
  // create a note with harmonics of the given key and duration
  {
    double freq = getFrequency(key);
    return tone(freq, duration, 1.0);
  }

  public static double[] tone(String key, double duration, double amplitude)
  {
    double freq = getFrequency(key);
    return tone(freq, duration, amplitude);
  }


  public static double[] tone(double freq, double duration)
  {  return tone(freq, duration, 1.0);  }


  public static double[] tone(double freq, double duration, double amplitude)
  {  return tone(freq, duration, amplitude, AudioTools::sinWave);  }



  public static double[] tone(double freq, double duration, double amplitude, 
                              BiFunction<Double, Integer, Double> waveFn)
  /* create a tone (using the wave function) of the given frequency (Hz), 
     for the given duration (secs) scaled to the given volume (amplitude)
  */
  {
    int n = (int) (AudioTools.SAMPLE_RATE * duration);
    double[] audio = new double[n+1];
    for (int i = 0; i <= n; i++) {
      audio[i] = amplitude * waveFn.apply(freq, i);
      // audio[i] = amplitude * sinWave(freq, i);
      //  amplitude * Math.sin(2*Math.PI * i * freq / AudioTools.SAMPLE_RATE);
      if(audio[i] > 1.0)
        audio[i] = 1.0;
      else if( audio[i] < -1.0)
        audio[i] = -1.0;
    }
    return audio;
  } /* tone */



  public static double[] loadTones(String fnm)
  /* The file format are lines of:
        <name> <filename>         // a single sound file
     and blank lines and comment lines.
  */
  {
    File file = new File(fnm);
    if (!file.isFile()) {
      System.out.println("Error: " + fnm + " is not a file");
      return null;
    }

    double[] tones = new double[0];
    try {
      System.out.println("Loading tones file: " + fnm);
      BufferedReader br = new BufferedReader( new FileReader(fnm));
      String line;
      while((line = br.readLine()) != null) {
        line = line.trim();
        if (line.length() == 0)  // blank line
          continue;
        if (line.startsWith("//"))   // comment
          continue;
        tones = concat(tones, getTone(line));
      }
      br.close();
    }
    catch (IOException e)
    {  System.out.println("  " + e);  }

    return tones;
  }  // end of loadTones()


  private static double[] getTone(String line)
  {
    String[] valStr = line.split("\\s+");
    if (valStr.length == 2) {
      if (valStr[0].equals("PAUSE"))
        return tone(0, Double.parseDouble(valStr[1]));    // quiet for this time
      else
        return tone(valStr[0], Double.parseDouble(valStr[1]));     // key, duration
    }
    else if (valStr.length == 1)
      return tone(valStr[0], 0.35);     // key, duration
    else {
      System.out.println("Wrong number of values ("
                         + valStr.length + ") in " + line);
      return new double[0];
    }
  }  // end of getTone()




  public static double getFrequency(String key)
  {
    Double freq = noteMap.get(key);
    if (freq == null) {
      System.out.println("Key not found: " + key + "; using A4");
      freq = noteMap.get("A4");
    }
    return freq;
  } /* getFrequency */


  public static double keyNum2Freq(int keyNumber)
  // https://en.wikipedia.org/wiki/Piano_key_frequencies
  {  return 440.0 * Math.pow(2, (keyNumber - 49.0)/12.0);  }



  public static double[] harmonic(String key, double duration)
  // create a note with harmonics of the given frequency and duration
  {
    double freq = getFrequency(key);
    double[] a  = tone(freq, duration);
    double[] hi = tone(2*freq, duration);  // an ocatve higher
    double[] lo = tone(freq/2, duration);  // an octave lower

    return average(a, hi, lo);
  }


  // ------------ wave functions -------------
  /* These functions can be passed to the tone() function to
     change the way that a tone is generated. The default is
     to use sinWave().
  */


  public static double sinWave(double freq, int i)
  {
    double time = (double)i/AudioTools.SAMPLE_RATE;
    double angle = 2*Math.PI * freq * time;
    return Math.sin(angle);
  } /* sinWave */


  public static double sawWave(double freq, int i)
  // Saw tooth wave
  // Characterized by rising edge followed by a sharp drop off.
  {
    double time = (double)i/AudioTools.SAMPLE_RATE;
    double step = freq*time;
    return 2.0 * (step - (float)Math.round(step));
  } /* sawWave */


  public static double triangleWave(double freq, int i)
  {
    double time = (double)i/AudioTools.SAMPLE_RATE;
    double step = freq*time;
    return 1.0 - (4.0 * Math.abs(Math.round(step) - step));
  } /* triangleWave */


  public static double squareWave(double freq, int i)
  // A square wave alternates between 1 and -1 at regular intervals
  {
    double time = (double)i/AudioTools.SAMPLE_RATE;
    double step = freq*time;
    return (Math.round(step)%2 == 0) ? 1 : -1;
  } /* squareWave */


  // ---------------- others ------------------

  public static void pause(int ms)
  {
    try {
      Thread.sleep(ms);
    }
    catch(InterruptedException ex) { }
  }  // end of pause()


  public static void stats(double[] audio)
  // Print statistics about the audio
  {
    if (audio == null)
      return;

    double total = 0.0;
    double max = 0.0;
    // Calculate the average level and the maximum (absolute) level
    for (int i = 0; i < audio.length; i++) {
      total += audio[i];
      double val = Math.abs(audio[i]);
      if (val > max)
        max = val;
    }
    double avg = total / (audio.length);

    // find out the percent <1% max level and <5% max level
    long below1 = 0;
    long below5 = 0;
    for (int i = 0; i < audio.length; i++) {
      double val = Math.abs(audio[i]);
      if (val < (0.05 * max)) {
        below5++;
        if (val < (0.01 * max))
          below1++;
      }
    }

    System.out.printf("Duration: %.2f secs\n",
            ((double) audio.length / (double) AudioTools.SAMPLE_RATE));
    System.out.printf("No. of samples: %d\n", audio.length);
    System.out.printf("Average level: %.3f\n", avg);
    System.out.printf("Max level: %.3f\n", max);
    System.out.printf("Below 1%% of max level: %.2f%%\n",
                      ((double) below1 / (double) audio.length*100.0));
    System.out.printf("Below 5%% of max level: %.2f%%\n",
                      ((double) below5 / (double) audio.length*100.0));
  } /* stats */



  // ----------------- drawing -------------------


  public static void draw(double[] audio)
  {  draw(audio, DRAW_FNM);  }


  public static void draw(double[] audio, String fnm)
  {
    System.out.println("Drawing audio into file: " + fnm);
    ArrayList<Line2D.Double> lines = genLines(audio);
    BufferedImage im = drawImage(lines);
    saveImage(im, fnm);
  } // end of draw()


  private static ArrayList<Line2D.Double> genLines(double[] audio)
  {
    if (audio == null) {
      System.out.println("Error: audio is null");
      return null;
    }

    double xStep = (double)IM_WIDTH / (double) audio.length;
    double yStep = (double)IM_HEIGHT/2;
    double yOrigin = (double)IM_HEIGHT/2;   // middle of image height

    ArrayList<Line2D.Double> lines = new ArrayList<>();

    double yPrev = yOrigin;
    for (int i = 0; i < audio.length; i++) {
      double x = i * xStep;
      double yCurr = yOrigin + audio[i]*yStep;
      lines.add(new Line2D.Double(x, yPrev, x, yCurr));
      yPrev = yCurr;
    }

    return lines;
  } /* genLines */



  private static BufferedImage drawImage(ArrayList<Line2D.Double> lines)
  {
    if (lines == null)
      return null;

    BufferedImage im = new BufferedImage(IM_WIDTH, IM_HEIGHT,
                                    BufferedImage.TYPE_INT_RGB);
    Graphics2D g2 = im.createGraphics();

    g2.setBackground(Color.WHITE);
    g2.fillRect(0, 0, IM_WIDTH, IM_HEIGHT);

    g2.setColor(Color.BLUE);
    for (int i = 0; i < lines.size(); i++)
      g2.draw((Line2D) lines.get(i));

    g2.dispose();

    return im;
  }  // end of drawImage()



  private static void saveImage(BufferedImage im, String fnm)
  {
    if (im == null)
      return;

    try {
      ImageIO.write(im, "png", new File(fnm));
      System.out.println("Written audio image to " + fnm);
    }
    catch (IOException e)
    {  System.out.println("Unable to save audio image to " + fnm);  }
  } /* saveImage */



  public static void mixerInfo()
  {
    System.out.println("\n----------- Mixer Information -----------");
    int idx = 1;
    Mixer.Info[] mixerInfos = AudioSystem.getMixerInfo();

    for (Mixer.Info info: mixerInfos) {
      Mixer mixer = AudioSystem.getMixer(info);

      Line.Info[] lineInfos = mixer.getSourceLineInfo();
      for (Line.Info lineInfo : lineInfos) {
        System.out.println("\nS" + idx + ". Source: " + info.getName());
        // System.out.println("-- " + lineInfo);
        showLine(mixer, lineInfo);
        idx++;
        System.out.println();
      }
    }

    System.out.println("-----------------");
    idx = 1;
    for (Mixer.Info info: mixerInfos) {
      Mixer mixer = AudioSystem.getMixer(info);

      Line.Info[] lineInfos = mixer.getTargetLineInfo();
      for (Line.Info lineInfo : lineInfos) {
        System.out.println("\nT" + idx + ". Target: " + mixer);
        // System.out.println("-- " + lineInfo);
        showLine(mixer, lineInfo);
        idx++;
        System.out.println();
      }
    }

    System.out.println("-------------------------");
  } /* mixerInfo */


  private static void showLine(Mixer mixer, Line.Info lineInfo)
  {
    try {
      Line line = mixer.getLine(lineInfo);
      if (!(line instanceof Clip))
        line.open();
      // System.out.println("  -- " + line);

      Control[] controls = line.getControls();
      if (controls.length > 0) {
        System.out.println("    Controls: ");
        for (int i=0; i < controls.length; i++)
          System.out.println("    " + (i+1) + ". " + controls[i]);
      }

      if (line instanceof DataLine) {
        DataLine dataLine = (DataLine)line;
        DataLine.Info dataInfo = (DataLine.Info)dataLine.getLineInfo();
        // System.out.println("    DataLine name: " + dataInfo.toString());
        AudioFormat[] audioFormats = dataInfo.getFormats();
        if (audioFormats.length > 0) {
          System.out.println("    Audio Formats: ");
          for (int i=0; i < audioFormats.length; i++)
            System.out.println("    " + (i+1) + ". " + audioFormats[i]);
        }

      } 
      else if (line instanceof Port) {
        Port port = (Port)line;
        Port.Info portInfo = (Port.Info)port.getLineInfo();
        System.out.println("    Port name: " + portInfo.toString());
      } 
      else
        System.out.println("  Unknown: " + line );
      line.close();
    }
    catch (LineUnavailableException e) {
      System.out.println("  -- Could not create line");
    }
  }  // end of showLine()




  // --------- audio array manipulation ------------
  /*
     Each sample is recorded as a double between -1.O and 1.O  
     in the array. A sample with value 0 represents silence.
  */


  public static double[] copy(double[] arr)
  {  return concat(arr);  }


  public static double[] concat(double[]... arrs)
  // concatenate many arrays into one
  {
    int totLen = 0;
    for (double[] arr: arrs)
      totLen += arr.length;

    double[] dest = null;
    int destPos = 0;

    for (double[] arr: arrs) {
      if (dest == null) {
        dest = Arrays.copyOf(arr, totLen);
        destPos = arr.length;
      }
      else {
        System.arraycopy(arr, 0, dest, destPos, arr.length);
        destPos += arr.length;
      }
    }
    return dest;
  }  // end of concat()



  public static double[] average(double[]... arrs)
  // return average of many arrays as one array
  {
    if (arrs == null)
      return null;

    int numArrs = arrs.length;
    if (numArrs == 1)
      return arrs[0];

    boolean areEqual = true;
    int minLength = arrs[0].length;
    for (int i=1; i < numArrs; i++) {
      int len = arrs[i].length;
      if (minLength != len) {
        areEqual = false;
        if (minLength > len)
          minLength = len;  
      }
    }

    if (!areEqual)
      System.out.println("array are not equal length; using minimum: " + minLength);

    // compute the average
    double[] avg = new double[minLength];
    for (int idx = 0; idx < minLength; idx++) {
      avg[idx] = 0;
      for (int i = 0; i < numArrs; i++)
        avg[idx] += arrs[i][idx]/numArrs;
    }
    return avg;
  } // end of average()



  public static double[] weightSum(double[] a, double[] b, double weight)
  // return weighted sum of two arrays
  {
    if ((a == null) || (b == null))
      return null;

    if (a.length != b.length) {
      System.out.println("array lengths not equal: " + a.length +
                            " /= " + b.length);
      return null;
    }

    if ((weight < 0) || (weight > 1)) {
      System.out.println("weight out of range 0 to 1; using 0.5");
      weight = 0.5;
    }

    // compute the weighted sum
    double[] c = new double[a.length];
    for (int i = 0; i < a.length; i++)
      c[i] = a[i]*weight + b[i]*(1.0-weight);
    return c;
  } // end of weightSum()



  public static double[] setVolume(double[] audio, double gain)
  {
    if (audio == null)
      return null;

    double[] result = new double[audio.length];
    for (int i = 0; i < audio.length; i++) {
      result[i] = audio[i] * gain;
      if(result[i] > 1.0)
        result[i] = 1.0;
      else if( result[i] < -1.0)
        result[i] = -1.0;
    }
    return result;
  } /* setVolume */



  public static double[] reverse(double[] audio)
  {
    if (audio == null)
      return null;

    double[] result = new double[audio.length];
    for (int i = 0; i < audio.length; i++)
      result[i] = audio[audio.length-1 - i];
    return result;
  }  // end of reverse()


  public static double[] stretch(double[] audio)
  // doubles the length of the audio by adding average samples
  {
    if (audio == null)
      return null;

    double[] result = new double[audio.length*2];
    int j = 0;
    for (int i = 0; i < audio.length-1; i++) {
      result[j] = audio[i];
      result[j+1] = (audio[i] + audio[i+1])/2;
      j +=2;
    }
    result[j] = audio[audio.length-1];
    result[j+1] = audio[audio.length-1];
    return result;
  }  // end of stretch()


  public static double[] half(double[] audio)
  // halves the length of the audio by averaging neighbouring samples
  {
    if (audio == null)
      return null;

    int halfLen;
    boolean isEven;
    if (audio.length%2 == 0)  // even
      halfLen = audio.length/2;
    else
      halfLen = (audio.length-1)/2;
       // odd, so drop last sample from halving

    double[] result = new double[halfLen];
    for (int i = 0; i < halfLen; i++)
      result[i] = (audio[i*2] + audio[i*2+1])/2;
    return result;
  }  // end of half()



  public static double[] noise(double[] audio)
  // adds noise to the audio
  {
    if (audio == null)
      return null;

    double[] result = new double[audio.length];
    for (int i = 0; i < audio.length; i++) {
      result[i] = audio[i] + randNoise();
      if(result[i] > 1.0)
        result[i] = 1.0;
      else if( result[i] < -1.0)
        result[i] = -1.0;
    }
    return result;
  }  // end of noise()


  private static double randNoise()
  // returns a random double in the range -MAX_NOISE to MAX_NOISE
  {  return (rand.nextDouble()*2 - 1.0)*MAX_NOISE;  }



  public static double[] setLoops(double[] audio, int numLoops)
  // Loop the audio a certain number of times
  {
    if (audio == null)
      return null;
    if (numLoops <= 0)
      return null;

    double[] result = new double[audio.length*numLoops];
    int offset = 0;
    for(int cnt = 0; cnt < numLoops; cnt++) {
      for (int i = 0; i < audio.length; i++)
        result[offset+i] = audio[i];
      offset += audio.length;
    }
    return result;
  } // end of setLoops()



  public static double[] merge(double[] audio1, double[] audio2)
  /* Merges two audios together and return the resulting buffer.
     Does a simple average of the samples from each audio buffer.
     If buffers are different lengths, the result will be as long as the
     shorter buffer.
  */
  {
    if ((audio1 == null) || (audio2 == null))
      return null;

    if (audio1.length == audio2.length)
      return AudioTools.weightSum(audio1, audio2, 0.5);
    else if (audio1.length > audio2.length) {  // audio1 is bigger
      System.out.println("First audio too long; being reduced to length of second");
      double[] temp1 = new double[audio2.length];
      for (int i = 0; i < audio2.length; i++)
        temp1[i] = audio1[i];
      return AudioTools.weightSum(temp1, audio2, 0.5);
    }
    else {   // audio2 is bigger
      System.out.println("Second audio too long; being reduced to length of first");
      double[] temp2 = new double[audio1.length];
      for (int i = 0; i < audio1.length; i++)
        temp2[i] = audio2[i];
      return AudioTools.weightSum(audio1, temp2, 0.5);
    }
  }  // end of merge()



  public static double[] cut(double[] audio, double start, double end)
  // Return the audio between the start and end times (in seconds).
  {
    if (audio == null)
      return null;

    double duration = (double) audio.length / (double) AudioTools.SAMPLE_RATE;
    System.out.printf("Audio duration: %.2f secs\n", duration);

    if (end <= start) {
      System.out.println("Error: start time (" + start +
                                       ") is >= end time (" +  end + ")");
      return audio;
    }

    if ((start < 0) || (start >= duration)) {
      System.out.println("Error: start time (" + start + ") is out of range");
      return audio;
    }

    if (end > duration) {
      System.out.println("Error: end time (" + end + ") is out of range");
      return audio;
    }

    int startIndex = (int)Math.round(start * AudioTools.SAMPLE_RATE);
    int endIndex = (int)Math.round(end * AudioTools.SAMPLE_RATE);
    System.out.println("Samples cut range: " + startIndex + " -- " + endIndex);

    double[] result = new double[endIndex - startIndex + 1];
    int idx = 0;
    for (int i = startIndex; i <= endIndex; i++) {
      result[idx] = audio[i];
      idx++;
    }
    return result;
  }  // end of cut()



  public static double[] pad(double[] audio, double start, double end)
  // Pad silence (in seconds) before and/or after the audio
  {
    if (audio == null)
      return null;

    if (start < 0) {
      System.out.println("Error: start duration is negative");
      return audio;
    }
    if (end < 0) {
      System.out.println("Error: end duration is negative");
      return audio;
    }

    int startLen = (int)Math.round(start * AudioTools.SAMPLE_RATE);
    int endLen = (int)Math.round(end * AudioTools.SAMPLE_RATE);

    double[] result = new double[startLen + audio.length + endLen];
    for (int i = 0; i < audio.length; i++)
      result[startLen+i] = audio[i];
    return result;
  }  // end of pad()



  public static double[] echo(double[] audio, double secs, double vol)
  /* Echo the audio by the specified no. of seconds at a volume less than 1.0
     An original sample at time t is echoed at (t+seconds)*vol
  */
  {
    if (audio == null)
      return null;

    if ((vol <= 0) || (vol >= 1)) {
      System.out.println("Volume must be between 0 and 1; using 0.5");
      vol = 0.5;
    }

    double duration = (double) audio.length / ((double) AudioTools.SAMPLE_RATE);
    if ((secs <= 0) || (secs > duration)) {
      System.out.println("seconds is out of range; using duration/5");
      secs = duration/5.0;
    }

    double[] echoAudio = AudioTools.setVolume(audio, vol);
    double[] delayedEcho = AudioTools.pad(echoAudio, secs, 0);

    double[] longOrig = AudioTools.pad(audio, 0, secs);

    return AudioTools.merge(longOrig, delayedEcho);
  }  // end of echo()



  //  --------------- test rig -------------------


  public static void main(String [] args)
  {
    if (args.length != 1)
      System.out.println("java AudioTools <fnm>");
    else {
      double[] audio = AudioTools.read(args[0]);
      AudioTools.play(audio);
    }
  } /* main */


}  // end of AudioTools class
