
// ChartAudio.java
// Andrew Davison, ad@fivedots.coe.psu.ac.th, Feb. 2022

/*
  A time series chart where audio data is 
  dynamically added over time.

  It utilizes JFreeChart (https://www.jfree.org/jfreechart/), 
  and the code is compiled and run
  with compile.bat and run.bat
*/


import java.awt.*;
import java.awt.event.*;
import javax.swing.*;
import java.util.*;
import java.util.concurrent.*;
import java.text.*;

import org.jfree.chart.*;
import org.jfree.chart.axis.*;
import org.jfree.chart.plot.*;
import org.jfree.chart.ui.*;

import org.jfree.data.time.*;
import org.jfree.data.xy.*;



public class ChartAudio 
               extends ApplicationFrame 
               implements ActionListener, Viewer

{
  private static final String CHART_TITLE = "Audio Data";
  private static final double TIME_RANGE = 30000.0;    // ms
             // data is visible on screen for a 30 sec period

  private static final int REFRESH_TIME = 2;   // ms
  private static final float Y_MINMAX = 1;    
                               // absolute range of y-axis data

  private static final Color PALE_GRAY = new Color(220,220,220);


  private TimeSeries series;

  private javax.swing.Timer timer = 
                    new javax.swing.Timer(REFRESH_TIME, this);  
  private Random random = new Random();

  private ConcurrentLinkedQueue<Double> samples = 
                       new ConcurrentLinkedQueue<>();

  private long startTime, prevTime;
  private int extraSamples = 0;
  private boolean isPlayingFinished = false;


  public ChartAudio()
  {
    super(CHART_TITLE);

    series = new TimeSeries(CHART_TITLE);
    TimeSeriesCollection dataset = new TimeSeriesCollection(series);
    JFreeChart chart = createChart(dataset);

    ChartPanel chartPanel = new ChartPanel(chart);
    chartPanel.setPreferredSize(new Dimension(900, 400));

    JPanel p = new JPanel(new BorderLayout());
    p.add(chartPanel, BorderLayout.CENTER);
    setContentPane(p);

    startTime = System.currentTimeMillis();
    prevTime = startTime;

    // timer.setInitialDelay(1000);
    timer.start();

    pack();
    UIUtils.centerFrameOnScreen(this);
    setVisible(true);
  }  // end of ChartAudio()



  private JFreeChart createChart(XYDataset dataset)
  {
    JFreeChart chart = ChartFactory.createTimeSeriesChart(
                                 CHART_TITLE, "mm:ss", "Amplitude", 
                                 dataset, true, true, false);
    chart.setBackgroundPaint(PALE_GRAY);
    chart.removeLegend();

    XYPlot plot = chart.getXYPlot();

    plot.setBackgroundPaint(Color.WHITE);
    plot.setDomainGridlinesVisible(true);
    plot.setDomainGridlinePaint(Color.LIGHT_GRAY);
    plot.setRangeGridlinesVisible(true);
    plot.setRangeGridlinePaint(Color.LIGHT_GRAY);

    ValueAxis xAxis = plot.getDomainAxis();
    xAxis.setAutoRange(true);
    xAxis.setFixedAutoRange(TIME_RANGE);
    // xAxis.setVerticalTickLabels(true);

    ((DateAxis)xAxis).setDateFormatOverride(
                              new SimpleDateFormat("mm:ss"));

    ValueAxis yAxis = plot.getRangeAxis();
    yAxis.setRange(-Y_MINMAX, Y_MINMAX);

    // highlight y == 0 line
    Marker yOrigin = new ValueMarker(0);
    yOrigin.setPaint(Color.BLACK);
    yOrigin.setStroke(new BasicStroke(1f));
    plot.addRangeMarker(yOrigin);

    return chart;
  } /* createChart */


  public void actionPerformed(ActionEvent e)
  // Generate an random entry
  {
    long currTime = System.currentTimeMillis();
    long playingTime = currTime - startTime;

    long interval = currTime - prevTime;
    prevTime = currTime;

    int numSamples = (int) Math.round(AudioTools.SAMPLE_RATE/1000.0 * interval) + extraSamples;

    double[] samps = getSamples(numSamples);
    if ((samps == null) && isPlayingFinished) {
      System.out.println("Drawing has finished");
      timer.stop();
      return;
    }

    if (samps == null)
      return;

    if (samps.length < numSamples)   // didn't get all samples
      extraSamples = numSamples - samps.length;
    else
      extraSamples = 0;

    if (samps.length == 0)
      return;


    // double currValue = randomValue();
    //series.add(new FixedMillisecond(playingTime), currValue);

    // double average = getAverage(samps);
    // series.add(new FixedMillisecond(playingTime), average);

    // double median = getMedian(samps);
    // series.add(new FixedMillisecond(playingTime), median);

    Arrays.sort(samps);
    series.add(new FixedMillisecond(playingTime), samps[0]);
    series.add(new FixedMillisecond(playingTime+1), samps[samps.length-1]);

    //System.out.printf("Time: %d; No. samples: %d\n", 
    //                            playingTime, samps.length);
  } /* actionPerformed */


/*
  private double getAverage(double[] arr)
  {
    double tot = 0;
    for (int i=0; i < arr.length; i++)
      tot += arr[i];
    return (tot/arr.length);
  }


  private double getMedian(double[] arr)
  {
    Arrays.sort(arr);
    if (arr.length % 2 == 0)
      return (arr[arr.length/2] + arr[arr.length/2-1])/2;
    else
      return arr[arr.length/2];
  }


  private double randomValue()
  {   return random.nextGaussian() * Y_MINMAX/3;   } 
*/


  private double[] getSamples(int n)
  // try to remove n samples from the queue
  {
    ArrayList<Double> sampsList = new ArrayList<>();
    while ((sampsList.size() < n) && (samples.peek() != null))
      sampsList.add( samples.remove());

    // System.out.println("Charting " + sampsList.size() + " samples");
    if (sampsList.size() == 0)
      return null;
    
    double[] audioFrag = new double[sampsList.size()];
    for (int i=0; i < sampsList.size(); i++)
      audioFrag[i] = sampsList.get(i);

    return audioFrag;
  }


  // Viewer methods ------------


  public void update(double[] audioFrag, long duration)
  {
    // System.out.println("Updated "  + audioFrag.length + " samples");
    for(double sample : audioFrag)
      samples.add(sample);
  }


  public void finished()
  {  
    System.out.println("All samples added");
    isPlayingFinished = true;
  }



  // ----------------------------------------

  public static void main(String[] args)
  {
    ChartAudio demo = new ChartAudio();   // used for debugging
  } 


}  // end of ChartAudio class
