
// ClipPlayer.java
// https://raw.githubusercontent.com/Hongten/mp3/master/src/test/com/b510/mp3/util/SoundPlayerTest.java
// Modified by Andrew Davison, Feb 2022, ad@fivedots.coe.psu.ac.th

/* 
   Load and play a sound clip, displaying a play/stop button, a progress bar, 
   and a volume control slider. 
*/


import java.io.*;
import java.util.*;

import javax.swing.*;
import javax.swing.event.*;
import javax.swing.border.*;
import java.awt.*;
import java.awt.event.*;

import javax.sound.sampled.*;



public class ClipPlayer extends JFrame
{
  private String fnm;
  private Clip clip; 
  private boolean isPlaying = false; 
  private int audioLen;       // length of the clip in ms
  private int audioPos = 0;   // current position within the clip in ms

  // GUI elements
  private JButton playButton; 
  private JSlider progressBar; 
  private JLabel timeLabel, statusLabel;

  private javax.swing.Timer timer;  // updates the progress bar



  public ClipPlayer(String fnm) 
  {
    super("ClipPlayer");

    this.fnm = fnm;
    clip = AudioTools.loadClip(fnm);
    if (clip == null)
      System.exit(1);

    audioLen = (int)(clip.getMicrosecondLength()/1000);
    initGUI();

    setDefaultCloseOperation( JFrame.EXIT_ON_CLOSE );
    pack();
    setResizable(false);    // fixed size display
    setVisible(true);
  } // end of ClipPlayer()



  private void initGUI()
  {
    Container c = getContentPane();
    c.setLayout( new BorderLayout());

    playButton = new JButton("Play");            // Play/stop button
    progressBar = new JSlider(0, audioLen, 0);   // Shows clip position

    timeLabel = new JLabel("0.0s", SwingConstants.RIGHT); // Shows time
    timeLabel.setPreferredSize(new Dimension(40,10));

    // start or stop playing the sound
    playButton.addActionListener(new ActionListener() {
      public void actionPerformed(ActionEvent e) {
        if (isPlaying)
          stop();
        else
          play();
      }
    });

    progressBar.addChangeListener(new ChangeListener() {
      public void stateChanged(ChangeEvent e) 
      {
        int value = progressBar.getValue(); // update the time label
        timeLabel.setText((value/1000 + "." + 
                          (value % 1000)/100) + "s");  // 1 dp
        // If we're not already there, jump there
        if (value != audioPos)
          jumpTo(value);
      }
    });

    // call tick() 10 times a second to keep
    // the progress bar in sync with the clip
    timer = new javax.swing.Timer(100, new ActionListener() {
      public void actionPerformed(ActionEvent e) 
      {  tick(); }
    });

    // put the controls in a row
    JPanel p1 = new JPanel();
    p1.setLayout( new BoxLayout(p1, BoxLayout.X_AXIS));
    p1.setBorder(new EmptyBorder(5, 5, 5, 5));
    p1.add(playButton);
    p1.add( Box.createRigidArea( new Dimension(5, 0)));
    p1.add(progressBar);
    p1.add(timeLabel);

    // the filename and duration
    JPanel p2 = new JPanel();
    double duration = (double)audioLen/1000.0;
    statusLabel = new JLabel(fnm + " (" +
                String.format("%.2f", duration) + "s)");
    p2.add(statusLabel);

    // add volume control if available
    JPanel p3 = null;
    try {
      FloatControl volControl = 
           (FloatControl) clip.getControl(FloatControl.Type.MASTER_GAIN);
      if (volControl != null) {
        p3 = new JPanel( new BorderLayout());
        p3.add( createSlider("Volume", volControl));
      }
    }
    catch (IllegalArgumentException e) {
      // If the MASTER_GAIN volume control is unsupported, just skip it
    }

    c.add(p2, BorderLayout.NORTH);
    c.add(p1, BorderLayout.CENTER);
    if (p3 != null)
      c.add(p3, BorderLayout.SOUTH);
  }  // end of initGUI()



  private JSlider createSlider(String title, FloatControl ctrl)
  // Return a JSlider to manipulate the FloatControl
  {
    if (ctrl == null)
      return null;

    JSlider slider = new JSlider(0, 1000);
    float min = ctrl.getMinimum();
    float max = ctrl.getMaximum();
    float width = max - min;

    slider.setValue((int) ((ctrl.getValue() - min) / width * 1000));

    Hashtable<Integer,JLabel> labels = new Hashtable<>(2);
    labels.put(0, new JLabel("Min"));
    labels.put(1000, new JLabel("Max"));

    slider.setLabelTable(labels);
    slider.setPaintLabels(true);
    //slider.setMajorTickSpacing(100);
    //slider.setPaintTicks(true);

    TitledBorder tb = new TitledBorder(title);
    tb.setTitlePosition(TitledBorder.BELOW_TOP);
    // tb.setTitleJustification(TitledBorder.CENTER);
    slider.setBorder(tb);

    slider.addChangeListener(new ChangeListener() {
      public void stateChanged(ChangeEvent e) 
      {
        float val = min + (slider.getValue()*width/1000.0f);
        ctrl.setValue(val);
      }
    });
    return slider;
  } /* createSlider */



  public void play()
  // Start playing the sound at the current pos
  {
    clip.start();
    timer.start();
    isPlaying = true;
    playButton.setText("Stop");
  } /* play */


  public void stop()
  // Stop playing the sound, but retain the current pos
  {
    timer.stop();
    clip.stop();
    isPlaying = false;
    playButton.setText("Play");
  } /* stop */


  public void reset()
  // Stop playing the sound and reset the pos to 0
  {
    stop();
    clip.setMicrosecondPosition(0);
    audioPos = 0;
    progressBar.setValue(0);
  } /* reset */


  public void jumpTo(int pos)   
  // jump to the specified position;
  // Called when the user drags the progress bar
  {
    if ((pos < 0) || (pos > audioLen))
      return;
    audioPos = pos;
    clip.setMicrosecondPosition(pos * 1000);
    progressBar.setValue(pos); // in case jumpTo() is called from outside
  } /* jumpTo */



  private void tick()
  // update the progress bar
  // The Timer object calls this method 10 times a second.
  // If the sound has finished, it resets to the beginning
  {
    if (clip.isActive()) {
      audioPos = (int) (clip.getMicrosecondPosition() / 1000);
      progressBar.setValue(audioPos);
    }
    else
      reset();
  } /* tick */



  // -----------------------------------------

  public static void main(String args[])
  {
    if (args.length != 1)
      System.out.println("Usage java ClipPlayer <fnm>");
    else
      new ClipPlayer(args[0]);
  } /* main */


}  // end of ClipPlayer class
