
// Mozart.java
// https://github.com/chindesaurus/Mozart-Waltzes  
// https://introcs.cs.princeton.edu/java/assignments/mozart.html
// Modified by Andrew Davison, ad@fivedots.coe.psu.ac.th, Feb. 2022

/*
  Mozart.java simulates Wolfgang Amadeus Mozart's Musikalisches Würfelspiel
   -- Mozart's musical dice game.  
 
  In Mozart's game, we compose a two-part waltz (minuet and trio) by joining together  
  randomly selected, pre-composed measures.  
 
  Every time "java Mozart" is executed, a new pseudorandomly generated piece of music
  is composed and performed, with the option of saving it as a .wav file.
  
  The minuet is 16 measures long, having 176 possible pre-composed measures.  
  The minuet[][] table determines which pre-composed measures to play.
  
  Roll two 'dice' for each of the 16 measures. The sum of the rolls is on
  the vertical axis; the measure of the minuet is on the horizontal axis.
  
  The trio is also 16 measures long, with 96 possible pre-composed measures.  
  To determine which one to play, roll one 'die' and use the trio[][] in the
  same way as minuet[][].
  
  Composing a minuet and trio in this manner allows for  
  11^16 * 6^16 = 1.3 * 10^29 different waltzes!

*/

import java.util.*;


public class Mozart
{

  // each element represents the measure of the minuet;
  // stored in M<integer>.wav
  private final static int[][] minuet = {
    {96, 22, 141, 41, 105, 122, 11, 30, 70, 121, 26, 9, 112, 49, 109, 14},
    {32, 6, 128, 63, 146, 46, 134, 81, 117, 39, 126, 56, 174, 18, 116, 83},
    {69, 95, 158, 13, 153, 55, 110, 24, 66, 139, 15, 132, 73, 58, 145, 79},
    {40, 17, 113, 85, 161, 2,159, 100, 90, 176, 7, 4, 67, 160, 52, 170},
    {148, 74, 163, 45, 80, 97, 36, 107, 25, 143, 64, 125, 76, 136, 1, 93},
    {104, 157, 27, 167, 154, 68, 118, 91, 138, 71, 150, 29, 101, 162, 23, 151},
    {152, 60, 171, 53, 99, 133, 21, 127, 16, 155, 57, 175, 43, 168, 89, 172},
    {119, 84, 114, 50, 140, 86, 169, 94, 120, 88, 48, 166, 51, 115, 72, 111},
    {98, 142, 42, 156, 75, 129, 62, 123, 65, 77, 19, 82, 137, 38, 149, 8},
    {3, 87, 165, 61, 135, 47, 147, 33, 102, 4, 31, 164, 44, 59, 173, 78},
    {54, 130, 10, 103, 28, 37, 106, 5, 35, 20, 108, 92, 12, 124, 44, 131}
  };

  // each element represents the measure of the trio;
  // stored in T<integer>.wav
  private final static int[][] trio = {
    {72, 6, 59, 25, 81, 41, 89, 13, 36, 5, 46, 79, 30, 95, 19, 66},
    {56, 82, 42, 74, 14, 7, 26, 71, 76, 20, 64, 84, 8, 35, 47, 88},
    {75, 39, 54, 1, 65, 43, 15, 80, 9, 34, 93, 48, 69, 58, 90, 21},
    {40, 73, 16, 68, 29, 55, 2, 61, 22, 67, 49, 77, 57, 87, 33, 10},
    {83, 3, 28, 53, 37, 17, 44, 70, 63, 85, 32, 96, 12, 23, 50, 91},
    {18, 45, 62, 38, 4, 27, 52, 94, 11, 92, 24, 86, 51, 60, 78, 31}
  };

  private static final String WALTZ_DIR = "mozart/";


  private static Random rand = new Random();



  public static void main(String[] args)
  {
    if ((args.length > 1)) {
      System.out.println("Usage: java Mozart [ <savefnm.wav> ]");
      return;
    }
    String saveFnm = (args.length == 1) ? args[0] : null;

    String[] fnms = selectParts();
    System.out.println("Building with: \n" + Arrays.toString(fnms) + "\n");

    // build the waltz
    double[] waltz = new double[0];
    for (String fnm : fnms) {
      double[] sample = AudioTools.read(fnm);
      waltz = AudioTools.concat(waltz, sample);
    }

    AudioTools.stats(waltz);
    AudioTools.play(waltz);

    if (saveFnm != null) {
      AudioTools.save(saveFnm, waltz);
      System.out.println("Saved in: " + saveFnm);
    }
  } /* main */



  private static String[] selectParts()
  /* Return a String array that contains the names of the
     files randomly chosen to construct the minuet and trio;
     the filenames are of the form:
          "M<integer>.wav" and "T<integer>.wav"
  */
  {
    int numMeasures = minuet[0].length;   
         // trio[][] should be same column length as minuet[][]

    String[] fnms = new String[2 * numMeasures];

    // add minuet filenames
    int numRows = minuet.length;
    for (int i = 0; i < numMeasures; i++)
      fnms[i] = WALTZ_DIR + "M" + minuet[rand.nextInt(numRows)][i] + ".wav";

    // add trio filenames
    numRows = trio.length;
    for (int i = 0; i < numMeasures; i++)
      fnms[i+numMeasures] =  WALTZ_DIR + "T" + trio[rand.nextInt(numRows)][i] + ".wav";

    return fnms;
  } /* selectParts */


}  // end of Mozart class
