
// TestEffects.java
// Andrew Davison, ad@fivedots.coe.psu.ac.th, Feb. 2022

/* A simple user interface for trying out various audio effects
   implemented in AudioTools.java.

   Call:
     java TestEffects 
   to get help on how to use the effects. 

   Also see
     effectsTest.bat 
   for a series of predefined calls 
*/

import java.util.*;


public class TestEffects
{

  public static void main(String [] args)
  {
    if (args.length < 2) {
      showHelp();
      return;
    }

    double[] audio = AudioTools.read(args[0]);
    String effect = args[1].toLowerCase();
    double[] result = null;
    String outFnm = "";

    System.out.println("Effect: " + effect);

    if (effect.equals("play"))
      AudioTools.play(audio);
    else if (effect.equals("stats"))
      AudioTools.stats(audio);
    else if (effect.equals("draw")) {
      if (args.length < 3)
        System.out.println("Usage: <inFnm> draw <outFnm>");
      else
        AudioTools.draw(audio, args[2]);
    }
    else if (effect.equals("vol")) {
      if (args.length < 4)
        System.out.println("Usage: <inFnm> vol <factor> <outFnm>");
      else {
        double vol = getDouble(args[2]);
        outFnm = args[3];
        result = AudioTools.setVolume(audio, vol);
      }
    }
    else if (effect.equals("rev")) {
      if (args.length < 3)
        System.out.println("Usage: <inFnm> rev <outFnm>");
      else {
        outFnm = args[2];
        result = AudioTools.reverse(audio);
      }
    }
    else if (effect.equals("noise")) {
      if (args.length < 3)
        System.out.println("Usage: <inFnm> noise <outFnm>");
      else {
        outFnm = args[2];
        result = AudioTools.noise(audio);
      }
    }
    else if (effect.equals("stretch")) {
      if (args.length < 3)
        System.out.println("Usage: <inFnm> stretch <outFnm>");
      else {
        outFnm = args[2];
        result = AudioTools.stretch(audio);
      }
    }
    else if (effect.equals("half")) {
      if (args.length < 3)
        System.out.println("Usage: <inFnm> half <outFnm>");
      else {
        outFnm = args[2];
        result = AudioTools.half(audio);
      }
    }
    else if (effect.equals("loop")) {
      if (args.length < 4)
        System.out.println("Usage: <inFnm> loop <times> <outFnm>");
      else {
        int numLoops = Integer.parseInt(args[2]);
        outFnm = args[3];
        result = AudioTools.setLoops(audio, numLoops);
      }
    }
    else if (effect.equals("merge")) {
      if (args.length < 4)
        System.out.println("Usage: <inFnm> merge <inFnm2> <outFnm>");
      else {
        String inFnm2 = args[2];
        outFnm = args[3];
        double[] audio2 = AudioTools.read(inFnm2);
        result = AudioTools.merge(audio, audio2);
      }
    }
    else if (effect.equals("cut")) {
      if (args.length < 5)
        System.out.println("Usage: <inFnm> cut <start> <end> <outFnm>");
      else {
        double start = getDouble(args[2]);
        double end   = getDouble(args[3]);
        outFnm = args[4];
        result = AudioTools.cut(audio, start, end);
      }
    }
    else if (effect.equals("pad")) {
      if (args.length < 5)
        System.out.println("Usage: <inFnm> pad <start> <end> <outFnm>");
      else {
        double start = getDouble(args[2]);
        double end  = getDouble(args[3]);
        outFnm = args[4];
        result = AudioTools.pad(audio, start, end);
      }
    }
    else if (effect.equals("echo")) {
      if (args.length < 5)
        System.out.println("Usage: <inFnm> echo <secs> <vol> <outFnm>");
      else {
        double secs = getDouble(args[2]);
        double vol  = getDouble(args[3]);
        outFnm = args[4];
        result = AudioTools.echo(audio, secs, vol);
      }
    }
    else {
      System.out.println("Unknown effect: " + effect);
      showHelp();
    }

    // Write the output file if one exists
    if ((result != null) && (outFnm.length() > 0)) {
      AudioTools.save(outFnm, result);
      System.out.println("Wrote " + result.length + " samples to " + outFnm);
    }
  } /* main */



  private static void showHelp()
  {
    System.out.println("Usage: java TestEffects <inFnm> <effect> [options]");
    System.out.println("<effect> [options] can be:");
    System.out.println(
      "  play    // play the file");
    System.out.println(
      "  stats   // print statistics about the audio");
    System.out.println(
      "  draw <outPNGFnm>   // draw the audio to a PNG file");
    System.out.println(
      "  stretch <outFnm>   // doubles the length of the audio");
    System.out.println(
      "  half <outFnm>   // halves the length of the audio");
    System.out.println(
      "  noise <outFnm>   // add noise to the audio");
    System.out.println(
      "  rev <outFnm>   // reverse the audio");
    System.out.println(
      "  vol <factor> <outFnm>     // change the volume by the given %factor");
    System.out.println(
      "  loop <times> <outFnm>     // loop audio the given number of times");
    System.out.println(
      "  merge <inFnm2> <outFnm>   // merge two audio files together ");
    System.out.println(
      "  cut <start> <end> <outFnm>   // clip audio between [start, end] seconds");
    System.out.println(
      "  pad <start> <end> <outFnm>   // add silence to start and/or end");
    System.out.println(
      "  echo <secs> <factor> <outFnm>   // add an echo");
    System.out.println();
    System.out.println("Supports only 44.1 kHz mono 16-bit signed WAV audio.");
  }  // end of showHelp()


  public static double getDouble(String s)
  { 
    double num = 0;
    try {
      num = Double.parseDouble(s);
    }
    catch (NumberFormatException ex){ 
      System.out.println("Incorrect format for double " + s); 
      System.exit(1);
    }
    return num;
  }  // end of getDouble()


}  // end of TestEffects class
